/***************************************************************************
                          vclib.c  -  description
                             -------------------
    begin                : Thu Jan 2 2003
    copyright            : (C) 2003 by Robert Penz
    email                : robert.penz@outertech.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <fcntl.h>
#include <errno.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/sysmacros.h>

#include <libgen.h>

#include <vclib.h>


/* presumes Line is preloaded with the "current" line of the file */
int cue_getTrackinfo(char *Line, tTrack *track, FILE *fdCueFile)
{
  char inum[3];
  char min;
  char sec;
  char fps;
  int already_set = 0;

  /* Get the 'mode' */
  if (strncmp(&Line[2], "TRACK ", 6)==0)
  {
/*    strncpy(track->num, &Line[8], 2); track->num[2] = '\0'; */

    track->mode = UNKNOWN;
    if(strncmp(&Line[11], "AUDIO", 5)==0) track->mode = AUDIO;
    if(strncmp(&Line[11], "MODE1/2352", 10)==0) track->mode = MODE1_2352;
    if(strncmp(&Line[11], "MODE1/2048", 10)==0) track->mode = MODE1_2048;
    if(strncmp(&Line[11], "MODE2/2352", 10)==0) track->mode = MODE2_2352;
    if(strncmp(&Line[11], "MODE2/2336", 10)==0) track->mode = MODE2_2336;
  }
  else return(1);

  /* Get the track indexes */
  while(1) {
    if(! fgets( Line, 256, fdCueFile ) ) { break;}

    if (strncmp(&Line[2], "TRACK ", 6)==0)
    {
      /* next track starting */
      break;
    }

    /* Track 0 or 1, take the first an get fill the values*/
    if (strncmp(&Line[4], "INDEX ", 6)==0)
    {
      /* check stuff here so if the answer is false the else stuff below won't be executed */
      strncpy(inum, &Line[10], 2); inum[2] = '\0';
      if ((already_set == 0) &&
          ((strcmp(inum, "00")==0) || (strcmp(inum, "01")==0)))
      {
        already_set = 1;

        min = ((Line[13]-'0')<<4) | (Line[14]-'0');
        sec = ((Line[16]-'0')<<4) | (Line[17]-'0');
        fps = ((Line[19]-'0')<<4) | (Line[20]-'0');

        track->minute = (((min>>4)*10) + (min&0xf));
        track->second = (((sec>>4)*10) + (sec&0xf));
        track->frame  = (((fps>>4)*10) + (fps&0xf));
      }
    }
    else if (strncmp(&Line[4], "PREGAP ", 7)==0) { ; /* ignore */ }
    else if (strncmp(&Line[4], "FLAGS ", 6)==0)  { ; /* ignore */ }
    else { printf("Unexpected cuefile line: %s\n", Line); }
  }
  return(0);
}


/* this function finds and opens the bin file */
int cue_find_bin (char *firstline, int *fd_bin, char *sBinFilename, char *sPath, char *sCueFilename)
{
  int i,j;
  char s[256];
  char t[256];

  /* get the filename out of that */
  /*                      12345 6  */
  if (strncmp(firstline, "FILE \"",6)==0)
  {
    i = 0;
    j = 0;
    while ( firstline[6 + i] != '"')
    {
      sBinFilename[j] = firstline[6 + i];

      /* if I found a path info, than delete all bevor it */
      switch (sBinFilename[j])
      {
        case '\\':
          j = 0;
          break;

        case '/':
          j = 0;
          break;

        default:
          j++;
      }
      i++;
    }
    sBinFilename[j] = '\0';

    /* printf ("<%s>\n", sBinFilename); */

  }


  /* now try to open that file, without path */
  *fd_bin = open (sBinFilename, O_RDONLY);
  if (*fd_bin == -1)
  {
    printf ("%s not found\n", sBinFilename);

    /* now try to find it with the path of the cue file */
    sprintf(s,"%s/%s",sPath, sBinFilename);
    *fd_bin = open (s, O_RDONLY);
    if (*fd_bin == -1)
    {
      printf ("%s not found\n", s);
      /* now I would say the whole filename is shit, build our own */
      strncpy(s, sCueFilename, strlen(sCueFilename) - 3 );
      s[strlen(sCueFilename) - 3] = '\0';
      strcat(s, "bin");
      *fd_bin = open (s, O_RDONLY);
      if (*fd_bin == -1)
      {
        printf ("%s not found\n", s);
        /* ok try it with path */
        sprintf(t,"%s/%s",sPath, s);
        *fd_bin = open (t, O_RDONLY);
        if (*fd_bin == -1)
        {
          printf ("%s not found\n", s);

          /* now I would say the whole filename is shit, build our own */
          strncpy(s, sCueFilename, strlen(sCueFilename) - 3 );
          s[strlen(sCueFilename) - 3] = '\0';
          strcat(s, "img");
          *fd_bin = open (s, O_RDONLY);
          if (*fd_bin == -1)
          {
            printf ("%s not found\n", s);
            /* ok try it with path */
            sprintf(t,"%s/%s",sPath, s);
            *fd_bin = open (t, O_RDONLY);
            if (*fd_bin == -1)
            {
              printf ("%s not found\n", s);

              /* I'll give up, maybe later I inlucde e.g. .img */
              printf ("couldn't find the bin file - giving up");
              return VC_COULD_NOT_OPEN_BIN_FILE;;
            }
          }
        } else strcpy(sBinFilename, t);

      } else strcpy(sBinFilename, s);

    } else strcpy(sBinFilename, s);

  }

  printf ("using %s\n", sBinFilename);

  return VC_ALL_OK;
}


static inline int cue_msf_2_sector(int minute, int second, int frame)
{
 return frame + (second + minute * 60 ) * 75;
}

static inline int cue_mode_2_sector_size(int mode)
{
  switch (mode)
  {
    case AUDIO:      return AUDIO;
    case MODE1_2352: return CD_FRAMESIZE_RAW;
    case MODE1_2048: return CD_FRAMESIZE;
    case MODE2_2352: return CD_FRAMESIZE_RAW;
    case MODE2_2336: return CD_FRAMESIZE_RAW0;

    default:
    	return 0;
  }

}

int cue_read_cue (char *cue_filename, struct virtualcd_info *vc_info)
{
  struct stat filestat;
  char sLine[256];
  unsigned int sect;
  char *s,*t;
  int i, error;
  FILE *fdCueFile;
  char sBinFilename[256];
  char sCueFilename[256];
  char sPath[256];

  /* we have no tracks at the beginning */
  vc_info->nTracks = 0;

  error = VC_ALL_OK;

  /* split the filename into a path and filename part */
  s = strdup(cue_filename);
  t = dirname(s);
  strcpy(sPath,t);

  /* printf ("path=%s\n", sPath); */

  /* no path at all? */
  if (strcmp(sPath, ".") == 0)
  {
    strcpy(sCueFilename,cue_filename);
  } else
  {
    strcpy(sCueFilename,cue_filename + strlen(sPath) + 1);
  }

  /* printf ("cue=%s\n", sCueFilename); */


  /* open the cue file */
  fdCueFile = fopen (cue_filename, "r");
  if (fdCueFile == NULL)
  {
    /* printf ("Unable to open %s\n", cue_filename); */
    return VC_COULD_NOT_OPEN_CUE_FILE;
  }

  /* read the first line and hand it to find_bin, which will
     test more than one possible name of the file */

  if(! fgets( sLine, 256, fdCueFile ) )
  {
    return VC_ERROR_READING_CUE_FILE;
  }

  error = cue_find_bin(sLine, &vc_info->ffd, sBinFilename, sPath, sCueFilename);
  if (error != VC_ALL_OK)
    return error;


  /* now build the track list */
  /* red the next line and call our track finder */
  if(! fgets( sLine, 256, fdCueFile ) )
  {
    return VC_ERROR_READING_CUE_FILE;
  }

  while(!feof(fdCueFile))
  {
    if (cue_getTrackinfo(sLine, &vc_info->tracks[vc_info->nTracks++], fdCueFile) != 0)
    {
    return VC_ERROR_READING_CUE_FILE;
    }
  }

  /* make a fake track with stands for the Lead out */
  if (fstat (vc_info->ffd, &filestat) == -1)
  {
    return VC_ERROR_STATTING_CUE_FILE;
  }
  sect = filestat.st_size / 2352;

  vc_info->tracks[vc_info->nTracks].frame = sect%75;
  sect=sect/75;
  vc_info->tracks[vc_info->nTracks].second = sect%60;
  sect=sect/60;
  vc_info->tracks[vc_info->nTracks].minute = sect;


  /* lets calculate the start sectors and offsets */
  for(i = 0; i <= vc_info->nTracks; i++)
  {
    vc_info->tracks[i].start_sector = cue_msf_2_sector(vc_info->tracks[i].minute,
                                      vc_info->tracks[i].second,
                                      vc_info->tracks[i].frame);

    /* if we're the first track we don't need to offset of the one befor */
    if (i == 0)
    {
      /* was always 0 on my svcds, but who knows */
      vc_info->tracks[0].start_offset = vc_info->tracks[0].start_sector
                                        * cue_mode_2_sector_size(vc_info->tracks[0].mode);
    } else
    {
      vc_info->tracks[i].start_offset = vc_info->tracks[i-1].start_offset +
                (vc_info->tracks[i].start_sector - vc_info->tracks[i-1].start_sector)
                * cue_mode_2_sector_size(vc_info->tracks[i-1].mode);
    }
  }

  /* only for testing print all tracks */
  for(i = 0; i < vc_info->nTracks; i++)
  {
    printf("track%d: fps=%d\n",i + 1, vc_info->tracks[i].frame);
    printf("track%d: sec=%d\n",i + 1, vc_info->tracks[i].second);
    printf("track%d: min=%d\n",i + 1, vc_info->tracks[i].minute);
    printf("track%d: startsector=%d\n",i + 1, vc_info->tracks[i].start_sector);
    printf("track%d: offset=%d\n",i + 1, vc_info->tracks[i].start_offset);
  }

  printf("leadout: fps=%d\n",vc_info->tracks[vc_info->nTracks].frame);
  printf("leadout: sec=%d\n",vc_info->tracks[vc_info->nTracks].second);
  printf("leadout: min=%d\n",vc_info->tracks[vc_info->nTracks].minute);
  printf("leadout: startsector=%d\n",vc_info->tracks[vc_info->nTracks].start_sector);
  printf("leadout: offset=%d\n",vc_info->tracks[vc_info->nTracks].start_offset);


  fclose (fdCueFile);

  strcpy(vc_info->comment,sBinFilename);

  return error;
}


/* ******************************************************************************* */
/* ******************************************************************************* */
/* ******************************************************************************* */
/* *******************************************************************************
                     function that are visible extern
   ******************************************************************************* */
/* ******************************************************************************* */
/* ******************************************************************************* */
/* ******************************************************************************* */

int load_virtualcd (const char *device, const char *file, struct virtualcd_info *vc_info)
{
  int fd;
  int error;

  /* first read the cue file and get all needed data from it */
  /* the bin file is opened in cue_find_bin */
  error = cue_read_cue (file, vc_info);
  if (error != VC_ALL_OK)
    return error;

  /* get an handle for the device */
  if ((fd = open (device, O_RDONLY)) < 0)
  {
    return VC_COULD_NOT_OPEN_DEVICE;
  }

  /* tell now the kernel stuff the data it needs */
  if (ioctl (fd, VIRTUALCD_LOADCD, (void *)vc_info) < 0) {
    return VC_IOCTL_ERROR;
  }


/* 	
 if (ioctl (fd, LOOP_SET_STATUS, &loopinfo) < 0) {
		(void) ioctl (fd, LOOP_CLR_FD, 0);
		perror ("ioctl: LOOP_SET_STATUS");
		return 1;
	}
*/

  /* close both handles - job is done */
  close (fd);
  close (vc_info->ffd);

  return error;
}

int unload_virtualcd (const char *device)
{
  int fd;

  if ((fd = open (device, O_RDONLY)) < 0)
  {
    return VC_COULD_NOT_DELETE_DEVICE;
  }
  if (ioctl (fd, VIRTUALCD_UNLOADCD, 0) < 0)
  {
    return VC_IOCTL_ERROR;
  }
  close (fd);

  return VC_ALL_OK;
}
