/***************************************************************************
                          virtualcd.c  -  description
                             -------------------
    begin                :  2oo2-12-29
    copyright            : (C) 2003 by Robert Penz
    email                : robert.penz@outertech.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

/*
 * Version 0.4 - 2oo3-04-08
 * removed sys_call_table use
 * fixed bug with devfsd
 * some extra error handling for the userspace program

 * Version 0.3 - 2oo3-01-11
 * added check if a "cd is loaded"
 * removed all warnings
 * code cleanup

 * Version 0.2 - 2oo3-01-12
 * support for proc fs
 * changed the vclib.c so it can be used by gui programs

 * Version 0.1 - 2oo3-01-10
 * first test release (data and raw mode work - but no checks, if you do something
 * wrong you'll kill the kernel ;-)
 *
 * gcc virtualcd.c -o virtualcd.o -c -I/usr/src/linux/include -O2 -D__KERNEL__ -DMODULE -Wall
 */

#define MAJOR_NR  63   /* LOCAL/EXPERIMENTAL US */
/* LOOP_MAJOR */

#define DEVICE_NR(device) MINOR(device)  /* has no partition bits */
#define DEVICE_NAME "virtualcd"              /* name for messaging */
#define DEVICE_INTR virtualcd_intrptr        /* pointer to bottom half */
#define DEVICE_NO_RANDOM                 /* no entropy to contribute */
#define DEVICE_REQUEST vc_request
#define DEVICE_OFF(d) /* do-nothing */


#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/fs.h>
#include <linux/file.h>
#include <linux/stat.h>
#include <linux/errno.h>
#include <linux/major.h>
#include <linux/wait.h>
#include <linux/blk.h>
#include <linux/blkpg.h>
#include <linux/init.h>
#include <linux/devfs_fs_kernel.h>
#include <linux/smp_lock.h>
#include <linux/swap.h>
#include <linux/slab.h>
#include <linux/proc_fs.h>
#include <sys/syscall.h>
#include <asm/uaccess.h>

#include <linux/loop.h>

#include <linux/kdev_t.h>
#include <linux/cdrom.h>


#define VC_DEBUG

#define AUDIO 0
#define MODE1 1
#define MODE2 2
#define MODE1_2352 10
#define MODE2_2352 20
#define MODE1_2048 30
#define MODE2_2336 40

#define VIRTUALCD_LOADCD		0x5400  /* load the virtual cd */
#define VIRTUALCD_UNLOADCD		0x5401  /* unload the virtual cd*/

/* *********************************************************************** */
/* should be moved to the header file */
/* *********************************************************************** */
typedef struct track
{
   unsigned short mode;
   unsigned short minute;
   unsigned short second;
   unsigned short frame;

   /* (min*60 + sec) * 75 + fps   */

   unsigned long start_sector;

   /* = the sizes in bytes off all tracks bevor this one */
   /* its needed if there are mode1 tracks befor the mpeg tracks */
   unsigned long start_offset;

   /*   unsigned char num[3]; */
} tTrack;

struct virtualcd_device {
	int			vc_number;
	int			vc_refcnt;
	kdev_t			vc_device;
/*	struct cdrom_tochdr	header; */
	tTrack			tracks[100];
	int 			nTracks; /* number of tracks on the cd */
	struct semaphore	vc_ctl_mutex;/* used for locking in vc_ioctl */
	struct file 		*vc_backing_file;
	int			loaded;
	char			comment[256];
};


struct virtualcd_info {
	tTrack			tracks[100];
	unsigned int		ffd;
	int 			nTracks; /* number of tracks on the cd */
	char			comment[256];
};

/* *********************************************************************** */
/* *********************************************************************** */
/* *********************************************************************** */

static int max_virtualcds = 8;

static struct virtualcd_device *virtualcd_dev;
static int *virtualcd_sizes;
static int *virtualcd_blksizes;

static devfs_handle_t devfs_handle;      /*  For the directory */
static char buf[CD_FRAMESIZE_RAW];	 /*separate buffer for the ioctls */

/* ************************* proc stuff ******************************** */

/* extern void * sys_call_table[]; */
/*static asmlinkage long (*old_open)(const char *filename, int flags, int mode);
*/
static struct proc_dir_entry* proc_entry;

/*static long new_open(const char *filename, int flags, int mode)
{
        return old_open(filename, flags, mode);
}*/

static int proc_read(char* buf, char** start, off_t offset, int count, int* eof, void* data)
{
        int len = 0;
	int i;
	len += snprintf(len+buf, 1024, "Drive Loaded Comment\n");
	for (i=0; i < max_virtualcds; i++)
	{
		if (virtualcd_dev[i].loaded == 0)
			len += snprintf(len+buf, 1024, "%3i:     %i   NO_CD_LOADED\n", i,
				virtualcd_dev[i].loaded);
		else
			len += snprintf(len+buf, 1024, "%3i:     %i   %s\n", i,
				virtualcd_dev[i].loaded, virtualcd_dev[i].comment);
	}
        *eof = 1;
        return len;
}


/* ********************** end proc stuff ******************************** */


static inline int vc_msf_2_sector(int minute, int second, int frame)
{
 return frame + (second + minute * 60 ) * 75;
}


static inline int vc_mode_2_sector_size(int mode)
{
  switch (mode)
  {
    case AUDIO:      return AUDIO;
    case MODE1_2352: return CD_FRAMESIZE_RAW;
    case MODE1_2048: return CD_FRAMESIZE;
    case MODE2_2352: return CD_FRAMESIZE_RAW;
    case MODE2_2336: return CD_FRAMESIZE_RAW0;

    default:
    	return 0;
  }

}


static int vc_findtrack(struct virtualcd_device *vc, int sector)
{
	int i;
	for(i = vc->nTracks - 1; i >= 0 ; i--)
		if (vc->tracks[i].start_sector <= sector)
			return i;
	return 0;
}

static struct virtualcd_device *vc_locate_device(const struct request *req)
{
	int dev;
	struct virtualcd_device *vc;

	/* Check if the minor number is in range */
	dev = DEVICE_NR(req->rq_dev);
	if (dev >= max_virtualcds) {
		static int count = 0;
		if (count++ < 5) /* print the message at most five times */
			printk(KERN_WARNING "virtualcd: request for unknown device\n");
		return NULL;
	}
	vc = &virtualcd_dev[dev]; /* Pick it out of device array */
	return vc;
}

int vc_read_actor(read_descriptor_t * desc, struct page *page, unsigned long offset, unsigned long size)
{
        char *kaddr;
        unsigned long count = desc->count;

        if (size > count)
                size = count;


/*	#ifdef VC_DEBUG
	printk(KERN_INFO "size %li\n", size);
	#endif
*/
        kaddr = kmap(page);
        memcpy(desc->buf, kaddr + offset, size);
        kunmap(page);

        desc->count = count - size;
        desc->written += size;
        desc->buf += size;
        return size;
}


static int vc_transfer(struct virtualcd_device *vc, const struct request *req)
{
	ssize_t retval;
	int startposition, track;
	int offset;
	read_descriptor_t desc;
	loff_t position;
	long sector = req->sector / 4;

	/* thats just testing code as I don't know really how it works */

	track = vc_findtrack(vc, sector);

	/* we know the track get the mode of it */
	switch (vc->tracks[track].mode)
	{
		case MODE1_2352:
			offset = CD_HEAD_SIZE + CD_SYNC_SIZE;
			break;
		case MODE1_2048:
			offset = 0;
			break;
		default:{
			static int count = 0;
			if (count++ < 5) /* print the message at most five times */
				printk(KERN_WARNING "virtualcd: normal read not possible for that track mode\n");
			return 0; }
	}

	/* get the mode of the bin file part and calc the positon */
	startposition = vc->tracks[track].start_offset + (sector -
			vc->tracks[track].start_sector)
			* vc_mode_2_sector_size(vc->tracks[track].mode);

/*	#ifdef VC_DEBUG
	printk(KERN_INFO "startposition %i - buffer %p\n", startposition, req->buffer);
	#endif
*/

	/* we only support read, and we told the kernel that, so we
	   won't check it here */

	position = startposition + offset;
	desc.written = 0;
	desc.count = 512 * req->current_nr_sectors;
	desc.buf = req->buffer;
	desc.error = 0;
	do_generic_file_read(vc->vc_backing_file, &position,
		&desc, vc_read_actor);

	retval = desc.written;

	if (!retval){
		retval = desc.error;
		#ifdef VC_DEBUG
		printk(KERN_INFO "retval error %i\n", retval);
		#endif
		return retval;
	}

        return 1;
}



void vc_request(request_queue_t *q)
{
	struct virtualcd_device *vc;
	int status;

/*	#ifdef VC_DEBUG
	printk(KERN_INFO "vc_request: start\n");
	#endif
*/
	while(1)
	{
		INIT_REQUEST;	/* returns when queue is empty */

/*		#ifdef VC_DEBUG
		printk(KERN_INFO "request %p: cmd %i sec %li (nr. %li)\n", CURRENT,
			CURRENT->cmd,
			CURRENT->sector,
			CURRENT->current_nr_sectors);
		#endif
*/
		/* check if its 4 * 512 which is 2048 byte */
		if (CURRENT->current_nr_sectors > 4)
		{
			printk(KERN_INFO "vc_request: current_nr_sectors(%li) <= 4 \n",
				CURRENT->current_nr_sectors);
			end_request(0);
			continue;
		}

		/* Which "device" are we using? */
		vc = vc_locate_device (CURRENT);
		if (vc == NULL)
		{
			end_request(0);
			continue;
		}
		/* check if we've an "cd loaded" */
		if (vc->loaded == 0)
		{
			end_request(-ENOMEDIUM);
			continue;
		}

		/* Perform the transfer and clean up. */
		down(&vc->vc_ctl_mutex);
		status = vc_transfer(vc, CURRENT);
		up(&vc->vc_ctl_mutex);
		end_request(status);
	}
/*	#ifdef VC_DEBUG
	printk(KERN_INFO "vc_request: end\n");
	#endif
*/
}


/* *********************************************************************** */
/*                        interface to the user space */
/* *********************************************************************** */
static int vc_ioctl(struct inode *ip, struct file *fp, unsigned int cmd,
		       unsigned long arg)
{
	struct virtualcd_device *vc;
	struct cdrom_tocentry entry;
	tTrack *aTrack;
	struct cdrom_msf msf;
	struct virtualcd_info vc_info;
	struct cdrom_tochdr tocHdr;

	int dev, err;
	loff_t position;
	int sector;
	int track;
	int tmp;

	#ifdef VC_DEBUG
	if (cmd != CDROMREADRAW)
		printk(KERN_INFO "vc_ioctl: start - command:%x\n", cmd);
	#endif

	/* check if there is an inode and check also the major number*/
	if (!ip)
		return -EINVAL;
	if (MAJOR(ip->i_rdev) != MAJOR_NR) {
		printk(KERN_WARNING "vc_ioctl: pseudo-major != %d\n",
		       MAJOR_NR);
		return -ENODEV;
	}

	/* check if the minor number is smaller than our max_virtualscds */
	dev = MINOR(ip->i_rdev);
	if (dev >= max_virtualcds)
		return -ENODEV;

	/* call it vc for easier use */
	vc = &virtualcd_dev[dev];

	/* lock it */
	down(&vc->vc_ctl_mutex);

	
	/* check if we've an "cd loaded" */
	if (vc->loaded == 0)
	{
		/* only 3 commands can work without cd */
		if (cmd != CDROMEJECT && cmd != CDROMEJECT_SW && cmd != VIRTUALCD_LOADCD)
		{
			/* unlock it */
			up(&vc->vc_ctl_mutex);


			#ifdef VC_DEBUG
			if (cmd != CDROMREADRAW)
				printk(KERN_INFO "vc_ioctl: end\n");
			#endif

			return -ENOMEDIUM;
		}

	}


	/* decide what to do, means what command was given to us */

	err = 0;

	switch (cmd) {
	case CDROMSTART:	/* Spin up the drive. Don't know, what to do,
				   so do nothing */
		break;
	case CDROMSTOP:	/* Spin down the drive */
		break;
	case CDROMPAUSE:	/* Pause the drive, don't support audio currently */
		/* if (aztAudioStatus != CDROM_AUDIO_PLAY) */
		err = -EINVAL;
		break;
	case CDROMRESUME:	/* Play it again, Sam */
		/* if (aztAudioStatus != CDROM_AUDIO_PAUSED) */
		err = -EINVAL;
		break;
	case CDROMMULTISESSION:	/*multisession support -- is not possible with cue */
		err = -EINVAL;
		break;
	case CDROMPLAYTRKIND:	/* Play a track.  This currently ignores index. */
				/* not supported */
		err = -EINVAL;
		break;
	case CDROMPLAYMSF:	/* Play starting at the given MSF address. */
				/* not supported */
		err = -EINVAL;
		break;
	case CDROMREADTOCHDR:	/* Read the table of contents header */
		tocHdr.cdth_trk0 = 1;
		tocHdr.cdth_trk1 = vc->nTracks;
		if (copy_to_user((void *) arg, &tocHdr,
			sizeof tocHdr))
			err = -EFAULT;
		break;
	case CDROMREADTOCENTRY:	/* Read an entry in the table of contents */

		if (copy_from_user(&entry, (void *) arg, sizeof entry))
{
			err = -EFAULT;
			break;
		}

		/* if ((!aztTocUpToDate) || aztDiskChanged)
			aztUpdateToc(); */

		/* lead out or in our boundies, everthing els is an erro*/
		if (entry.cdte_track == CDROM_LEADOUT)
			aTrack = &(vc->tracks[vc->nTracks]);
		else if (entry.cdte_track > vc->nTracks
			 || entry.cdte_track < 1) {
			err = -EINVAL;
			printk(KERN_WARNING "vc_ioctl: there is no track %i\n", entry.cdte_track);
			break;
		} else /* normal case */
			aTrack = &(vc->tracks[entry.cdte_track]) - 1;

		/* entry.cdte_adr = tocPtr->ctrl_addr;
		entry.cdte_ctrl = tocPtr->ctrl_addr >> 4; */
		if (entry.cdte_format == CDROM_LBA)
			entry.cdte_addr.lba = 0;
			    /* azt_msf2hsg(&tocPtr->diskTime); */
		else if (entry.cdte_format == CDROM_MSF) {
			entry.cdte_addr.msf.minute = aTrack->minute;
			entry.cdte_addr.msf.second = aTrack->second;
			entry.cdte_addr.msf.frame  = aTrack->frame;
		} else {
			err = -EINVAL;
			printk(KERN_WARNING "vc_ioctl: there is no format %x\n", entry.cdte_format);
			break;
		}
		/* copy the data back into user space */
		if (copy_to_user((void *) arg, &entry, sizeof entry))
			err = -EFAULT;
		break;
	case CDROMSUBCHNL:	/* Get subchannel info - don't know what it does */
		err = -EIO;
		break;
	case CDROMVOLCTRL:	/* Volume control
					just ignore */
		break;
	case CDROMEJECT:	/* just ignore */
		break;
	case CDROMEJECT_SW:
		break;
	case CDROMRESET:
		break;
	case CDROMREADCOOKED:	/*read data in mode 1 (2048 Bytes) */
	case CDROMREADRAW:	/*read data in mode 2 (2336 Bytes) */
		if (copy_from_user(&msf, (void *) arg, sizeof msf)) {
			err = -EFAULT;
			break;
		}

		sector =vc_msf_2_sector(msf.cdmsf_min0, msf.cdmsf_sec0, msf.cdmsf_frame0);
		track = vc_findtrack(vc, sector);

		/* get the mode of the bin file part and calc the positon */
		position = vc->tracks[track].start_offset + (sector -
				vc->tracks[track].start_sector)
				* vc_mode_2_sector_size(vc->tracks[track].mode);

		#ifdef VC_DEBUG
		printk(KERN_INFO "sector:%i track:%i position %i\n", sector,track, (int) position);
		#endif

 		/* check if the track is at its end*/
		if (position >= vc->tracks[track+1].start_offset)
		{
			printk(KERN_WARNING "vc_ioctl: position out of track - sector:%i track:%i position %i\n",
				sector,track, (int) position);
			err = -EFAULT;
			break;
		}
		
		if (cmd == CDROMREADRAW) {
			tmp = generic_file_read(vc->vc_backing_file,(void *) arg,
				CD_FRAMESIZE_RAW, &position);

			if (tmp != CD_FRAMESIZE_RAW)
				printk(KERN_INFO "only read %i bytes\n",tmp);
		} else
		{	/*CDROMREADCOOKED*/
			tmp = generic_file_read(vc->vc_backing_file,(void *) arg,
				CD_FRAMESIZE, &position);

			if (tmp != CD_FRAMESIZE)
				printk(KERN_INFO "only read %i bytes\n",tmp);
		};

		break;
	case CDROMSEEK:	/*seek msf address - not supported */
		break;
	case CDROMREADMODE1:	/*set read data in mode 1 */
		/* return aztSetDiskType(AZT_MODE_1); */
		break;
	case CDROMREADMODE2:	/*set read data in mode 2 */
		/* return aztSetDiskType(AZT_MODE_2); */
		break;
	case VIRTUALCD_LOADCD:
		/* just update our virtual cd data */
		if (copy_from_user(&vc_info, (void *) arg, sizeof vc_info))
		{
			err = -EFAULT;
			break;
		}


		#ifdef VC_DEBUG
		printk(KERN_INFO "vc_ioctl: vc_info.ffd:%x\n", vc_info.ffd);
		#endif

		vc->vc_backing_file = fget(vc_info.ffd);
		if (!vc->vc_backing_file)
		{
			err = -EBADF;
			break;
		}

		memcpy(vc->tracks, vc_info.tracks, sizeof vc_info.tracks);
		vc->nTracks = vc_info.nTracks;
		strcpy(vc->comment, vc_info.comment);

		/* hard coded to get the size of the first track */
		virtualcd_sizes[dev] = vc->tracks[1].start_sector * CD_FRAMESIZE;

		/* open the bin file - at least inc the counter ;-)*/
		get_file(vc->vc_backing_file);

		vc->loaded = 1;

		MOD_INC_USE_COUNT;
		break;

	case VIRTUALCD_UNLOADCD:
		vc->loaded = 0;
		fput(vc->vc_backing_file);
		/* just close the bin file */
		MOD_DEC_USE_COUNT;
		break;
	default:
		err = -EINVAL;
	}

	/* unlock it */
	up(&vc->vc_ctl_mutex);

	#ifdef VC_DEBUG
	if (cmd != CDROMREADRAW)
		printk(KERN_INFO "vc_ioctl: end\n");
	#endif

	return err;
}


static int vc_open(struct inode *inode, struct file *file)
{
	struct virtualcd_device *vc;
	int	dev;

	#ifdef VC_DEBUG
	printk(KERN_INFO "vc_open: start\n");
	#endif

	/* check if there is an inode and check also the major number*/
	if (!inode)
		return -EINVAL;
	if (MAJOR(inode->i_rdev) != MAJOR_NR) {
		printk(KERN_WARNING "vc_open: pseudo-major != %d\n", MAJOR_NR);
		return -ENODEV;
	}

	/* check if the minor number is smaller than our max_virtualscds */
	dev = MINOR(inode->i_rdev);
	if (dev >= max_virtualcds)
		return -ENODEV;

	/* call it vc for easier use */
	vc = &virtualcd_dev[dev];

	/* one use more and lock */
	MOD_INC_USE_COUNT;
	down(&vc->vc_ctl_mutex);

	/* one ref more and unlock it */
	vc->vc_refcnt++;
	up(&vc->vc_ctl_mutex);

	#ifdef VC_DEBUG
	printk(KERN_INFO "vc_open: end\n");
	#endif

	return 0;
}

static int vc_release(struct inode *inode, struct file *file)
{
	struct virtualcd_device *vc;
	int	dev/*, type*/;

	#ifdef VC_DEBUG
	printk(KERN_INFO "vc_release: start\n");
	#endif


	/* check if there is an inode and check also the major number*/
	if (!inode)
		return 0;
	if (MAJOR(inode->i_rdev) != MAJOR_NR) {
		printk(KERN_WARNING "vc_release: pseudo-major != %d\n",
		       MAJOR_NR);
		return 0;
	}

	/* check if the minor number is smaller than our max_virtualscds */
	dev = MINOR(inode->i_rdev);
	if (dev >= max_virtualcds)
		return -ENODEV;

	/* call it vc for easier use */
	vc = &virtualcd_dev[dev];

	#ifdef VC_DEBUG
	printk(KERN_INFO "vc_release: before mutex\n");
	#endif

	/* lock it */
	down(&vc->vc_ctl_mutex);

	#ifdef VC_DEBUG
	printk(KERN_INFO "vc_release: after mutex\n");
	#endif


	/* one less and unlock it */
	--vc->vc_refcnt;

	up(&vc->vc_ctl_mutex);

	MOD_DEC_USE_COUNT;

	#ifdef VC_DEBUG
	printk(KERN_INFO "vc_release: end\n");
	#endif

	return 0;
}

static struct block_device_operations vc_fops = {
	owner:		THIS_MODULE,
	open:		vc_open,
	release:	vc_release,
	ioctl:		vc_ioctl,
};

/*
 * And now the modules code and kernel interface.
 */
MODULE_AUTHOR("robert.penz@outertech.com");
MODULE_PARM(max_virtualcds, "i");
MODULE_PARM_DESC(max_loop, "Maximum number of virtual cdroms (1-256)");
MODULE_LICENSE("GPL");

int __init virtualcd_init(void)
{
	int	i;

	#ifdef VC_DEBUG
	printk(KERN_INFO "virtualcd_init: start\n");
	printk(KERN_INFO "buf address:%p\n", buf);
	#endif


	/* check if the kernelparmeter is correct and register the devfs stuff */
	if ((max_virtualcds < 1) || (max_virtualcds > 256)) {
		printk(KERN_WARNING "virtualcd: invalid max_virtualcds (must be between"
				    " 1 and 256), using default (8)\n");
		max_virtualcds = 8;
	}

	if (devfs_register_blkdev(MAJOR_NR, "virtualcd", &vc_fops)) {
		printk(KERN_WARNING "Unable to get major number %d for virtualcd"
				    " device\n", MAJOR_NR);
		return -EIO;
	}

	devfs_handle = devfs_mk_dir(NULL, "virtualcd", NULL);
	devfs_register_series(devfs_handle, "%u", max_virtualcds, DEVFS_FL_DEFAULT,
			      MAJOR_NR, 0,
			      S_IFBLK | S_IRUSR | S_IWUSR | S_IRGRP,
			      &vc_fops, NULL);

	/* allocate max_virtualcds entries in our struct */
	virtualcd_dev = kmalloc(max_virtualcds * sizeof(struct virtualcd_device), GFP_KERNEL);
	if (!virtualcd_dev)
		return -ENOMEM;

	/* create the arrays for size and block size */
	virtualcd_sizes = kmalloc(max_virtualcds * sizeof(int), GFP_KERNEL);
	if (!virtualcd_sizes)
		goto out_sizes;

	virtualcd_blksizes = kmalloc(max_virtualcds * sizeof(int), GFP_KERNEL);
	if (!virtualcd_blksizes)
		goto out_blksizes;

	/* tell the kernel my read/write function */
	blk_init_queue(BLK_DEFAULT_QUEUE(MAJOR_NR), vc_request);


	for (i = 0; i < max_virtualcds; i++) {
		struct virtualcd_device *vc = &virtualcd_dev[i];
		memset(vc, 0, sizeof(struct virtualcd_device));
		init_MUTEX(&vc->vc_ctl_mutex);
			/*init_MUTEX_LOCKED(&lo->lo_sem);
			init_MUTEX_LOCKED(&lo->lo_bh_mutex); */
		vc->vc_number = i;
			/*spin_lock_init(&lo->lo_lock); */
	}

	memset(virtualcd_sizes, 0, max_virtualcds * sizeof(int));
/*	memset(virtualcd_blksizes, 0, max_virtualcds * sizeof(int)); */

	for (i=0; i < max_virtualcds; i++) /* all the same blocksize */
		virtualcd_blksizes[i] = CD_FRAMESIZE;

	blk_size[MAJOR_NR] = virtualcd_sizes;
	blksize_size[MAJOR_NR] = virtualcd_blksizes;

	for (i = 0; i < max_virtualcds; i++)
		register_disk(NULL, MKDEV(MAJOR_NR, i), 1, &vc_fops, 0);

	/* setup the proc stuff */
/*	old_open = sys_call_table[__NR_open];
        sys_call_table[__NR_open] = new_open; */

        proc_entry = create_proc_entry("virtualcd", 0440, NULL);

        if (proc_entry)
                proc_entry->read_proc = proc_read;
        else {
                printk("<1>Error registering /proc entry!\n");
                return -ENOENT;
        }

	/* all done report our success */
	printk(KERN_INFO "virtualcd: loaded (max %d devices)\n", max_virtualcds);

	#ifdef VC_DEBUG
	printk(KERN_INFO "virtualcd_init: end\n");
	#endif
	return 0;

out_sizes:
	kfree(virtualcd_dev);
out_blksizes:
	kfree(virtualcd_sizes);
	printk(KERN_ERR "virtualcd: ran out of memory\n");
	return -ENOMEM;
}

void virtualcd_exit(void)
{
	int i;

	#ifdef VC_DEBUG
	printk(KERN_INFO "virtualcd_exit: start\n");
	#endif

	/* get rid of the proc stuff */
/*        sys_call_table[__NR_open] = old_open; */
        if (proc_entry)
                remove_proc_entry("virtualcd", NULL);


	/* free all reference the kernel has in various caches */
	for (i = 0; i < max_virtualcds; i++)
		fsync_dev(MKDEV(MAJOR_NR, i)); /* flush the devices */

	/* Fix up the request queue(s) */
	blk_cleanup_queue(BLK_DEFAULT_QUEUE(MAJOR_NR));

	devfs_unregister(devfs_handle);
	if (devfs_unregister_blkdev(MAJOR_NR, "virtualcd"))
		printk(KERN_WARNING "virtualcd: cannot unregister blkdev\n");

	kfree(virtualcd_dev);
	kfree(virtualcd_sizes);
	kfree(virtualcd_blksizes);

	#ifdef VC_DEBUG
	printk(KERN_INFO "virtualcd_exit: end\n");
	#endif

}

module_init(virtualcd_init);
module_exit(virtualcd_exit);

#ifndef MODULE
static int __init max_virtualcds_setup(char *str)
{
	max_virtualcds = simple_strtol(str, NULL, 0);
	return 1;
}

__setup("max_virtualcds=", max_virtualcds_setup);
#endif

